package de.uks.plm.ftglb.incremental.bcg;

import x10.compiler.*;
import x10.util.Team;

import de.uks.plm.ftglb.incremental.TaskQueue;
import de.uks.plm.ftglb.incremental.TaskBag;
import de.uks.plm.ftglb.incremental.GLBResult;

public final class Queue extends BC implements TaskQueue[Queue, Double]
{
    public var lower:Rail[Int];
    public var upper:Rail[Int];
    protected var size:Long;
    public var state:Int = 0n;
    public var s:Int;

    public def this(rmat:Rmat, permute:Int) : Queue
    {
        super(rmat, permute);
        lower = new Rail[Int](4096);
        upper = new Rail[Int](4096);
        val h = here.id as Int;
        val max = Place.numPlaces();
        lower(0) = (N as Long * h/max) as Int;
        upper(0) = (N as Long * (h + 1) / max) as Int;
        size = 1;
    }

    public final def grow():void
    {
        val capacity = size * 2;
        val l = new Rail[Int](capacity);
        Rail.copy(lower, 0, l, 0, size);
        lower = l;
        val u = new Rail[Int](capacity);
        Rail.copy(upper, 0, u, 0, size);
        upper = u;
    }

    @Inline public def process(n:Long)
    {
        var i:Long = 0;
        val top = this.size - 1;
        val l = this.lower(top);
        val u = this.upper(top) - 1n;

        switch (state)
        {
            case 0n:
                refTime = System.nanoTime();
                s = this.verticesToWorkOn(u);
                this.state = 1n;

            case 1n:
                this.bfsShortestPath1(s);
                this.state = 2n;

            case 2n:
                while(!regularQueue.isEmpty())
                {
                    if (i++ > n)
                    {
                        return (true);
                    }
                    this.bfsShortestPath2();
                }
                this.state = 3n;

            case 3n:
                this.bfsShortestPath3();
                state = 4n;

            case 4n:
                while(!regularQueue.isEmpty())
                {
                    if (i++ > n)
                    {
                        return (true);
                    }
                    this.bfsShortestPath4(s);
                }
                this.accTime += ((System.nanoTime() - refTime) / 1e9);
                this.state = 0n;
                if(u == l)
                {
                  --this.size; 
                }
                else
                {
                    this.upper(top) = u;
                }
                this.protResults();
        }
        return (0 < this.size);
    }

    @Inline private def protResults() : void
    {
        this.betweennessMap = new Rail[Double](this.realBetweennessMap);
    }

    @Inline public def split()
    {
        var s:Long = 0;
        for (var i:Long = 0; i < this.size; ++i)
        {
            if (2 <= (this.upper(i) - this.lower(i)))
            {
                ++s;
            }
        }

        if (0 == s)
        {
            return (null);
        }

        val bag = new Bag(s);
        s = 0;
        for (var i:Long = 0; i < size; ++i)
        {
            val p = this.upper(i) - this.lower(i);
            if (2n <= p)
            {
                // bag.upper(s) = upper(i);
                // upper(i) -= (p / 2n);
                // bag.lower(s++) = upper(i);
               bag.lower(s) = this.lower(i);
               bag.upper(s) = this.upper(i) - ((p + 1n) / 2n);
               this.lower(i) = bag.upper(s++);
            }
        }
        return (bag);
    }

    @Inline public def merge(bag:Bag) : void
    {
        val bagSize = bag.size();
        val thisSize = size;
        while (upper.size < bagSize + thisSize)
        {
            grow();
        }
        Rail.copy(this.lower, 0, this.lower, bagSize, thisSize);
        Rail.copy(this.upper, 0, this.upper, bagSize, thisSize);

        Rail.copy(bag.lower, 0, this.lower, 0, bagSize);
        Rail.copy(bag.upper, 0, this.upper, 0, bagSize);
        size += bagSize;
    }

    @Inline public def merge(bag:TaskBag) {
        merge(bag as Bag);
    }

    @Inline public def merge(val that:TaskQueue[Queue, Double]) : void
    {
        this.merge(that as Queue);
    }

    @Inline public def merge(val that:Queue) : void
    {
        val thisSize:Long = this.size();
        val thatSize:Long = (null == that ? 0 : that.size());
        if (0 == thatSize)
        {
            return;
        }

        while (thisSize + thatSize > lower.size)
        {
            grow();
        }
        Rail.copy(that.lower, 0, lower, thisSize, thatSize);
        Rail.copy(that.upper, 0, upper, thisSize, thatSize);
        size += thatSize;
    }

    @Inline public def mergeResult(val that:TaskQueue[Queue, Double]) : void
    {
        this.mergeResult(that as Queue);
    }

    @Inline public def mergeResult(val that:Queue) : void
    {
        for (var idx:Long = 0; that.betweennessMap.size < idx; ++idx)
        {
            this.betweennessMap(idx) += that.betweennessMap(idx);
            this.realBetweennessMap(idx) += that.betweennessMap(idx);
        }
    }

    @Inline public def count() : Long
    {
        return (this.count);
    }

    @Inline public def size() : Long
    {
        return (this.size);
    }

    // @override
    public def getResult():BCGLBResult
    {
        
        val result = new BCGLBResult();
        
        return result;
    }

    public class BCGLBResult extends GLBResult[Double]
    {
        public def this(){}

        public  def getResult(): Rail[Double]
        {
            return (betweennessMap);
        }

        public  def getReduceOperator():Int
        {
            return Team.ADD;
        }

        public def display(r:Rail[Double]) : void
        {
            for(var i:Int=0n; i<N; ++i) {
                if(r(i) != 0.0) {
                    Console.OUT.println("(" + i + ") -> " + sub(""+r(i), 0n, 6n));
                }
            }
        }
    }

    /**
     * Override parent class method to print allocation and processing time
     */
    public def printLog():void{
        Console.OUT.println("[" + here.id + "]"
                + " Time = " + accTime
                + " Count = " + count);
        
        
    }
}
