package de.uks.plm.ftglb.incremental.dataStructures;

import x10.util.ArrayList;
import x10.util.List;

import de.uks.plm.ftglb.incremental.TaskQueue;

/**
 * This class describes the backup-data. These data are send to the backup-
 * place, which - in return - handles them according to the given backup-cause
 * and the given backup-type.
 */
public class BackupData[Queue, R] {Queue<:TaskQueue[Queue, R]}
{
    // public static val KIND_INCR:Long = 1;
    public static val KIND_COMP:Long = 2;

    
    public static val CAUSE_REGULAR:Long = 1;
    public static val CAUSE_STEAL:Long = 2;
    public static val CAUSE_TAKEOVER:Long = 4;
    public static val CAUSE_IA:Long = 8;

    /** The original uf of the source, from which the data originates. */
    public val sourceId:Long;

    /** Added Tasks since the last backup. */
    public val backup:Queue;

    /** The new log (if needed) */
    public val log:Long;

    /** Amount of removed tasks from the top since the last backup. */
    public val removeTop:Long;

    /** Amount of removed tasks from the bottom since the last backup. */
    public val removeBottom:Long;

    /** kind of the backup (full or incremental) */
    public val kind:Long;

    /** cause of the backup */
    public val cause:Long;

    /** if it is a steal-back, the tan is stored here */
    public val tan:Long;

    /** if it is a steal-back, this is the list of thieves */
    public val thieves:List[Long];

    /**
     * Constructor.
     *
     * @param sourceId the id of the place, where this data originated from.
     * @param removeTop number of tasks to pop from the last backup.
     * @param backup the tasks, as well as the partial result to add to the
     *        old backup.
     */
    // public def this(val sourceId:Long
    //                 , val removeTop:Long
    //                 , val backup:Queue) : BackupData[Queue, R]
    // {
    //     this(Long.MIN_VALUE
    //          , sourceId
    //          , KIND_INCR
    //          , CAUSE_REGULAR
    //          , removeTop
    //          , 0
    //          , -1
    //          , backup
    //          , null);
    // }

    // /**
    //  * Constructor
    //  *
    //  * @param removeTop number of tasks to pop from the last backup.
    //  * @param tan the tan for this transaction.
    //  */
    // public def this(val tan:Long
    //                 , val me:Long
    //                 , val removeTop:Long
    //                 , val thieves:List[Long]) : BackupData[Queue, R]
    // {
    //     this(tan
    //          , me
    //          , BackupData.KIND_INCR
    //          , BackupData.CAUSE_STEAL
    //          , removeTop
    //          , 0
    //          , -1
    //          , (null as Queue)
    //          , thieves);
    // }

    /**
     * Constructor.
     *
     * @param backup the tasks, as well as the partial result to add to the
     *        old backup.
     * @param tan the tan for this transaction.
     */
    public def this(val tan:Long
                    , val me:Long
                    , val backup:Queue
                    , val thieves:List[Long]) : BackupData[Queue, R]
    {
        this(tan
             , me
             , BackupData.KIND_COMP
             , BackupData.CAUSE_STEAL
             , 0
             , 0
             , -1
             , backup
             , thieves);
    }

    public def this(val me:Long, val backup:Queue)
    {
        this(Long.MAX_VALUE
             , me
             , BackupData.KIND_COMP
             , BackupData.CAUSE_REGULAR
             , 0
             , 0
             , -1
             , backup
             , null);
    }

    /**
     * Constructor.
     *
     * @param tan the tan for this transaction.
     * @param sourceId the id of the place, where this data originated from.
     * @param kind the kind of the backup.
     * @param cause the action, that caused the backup
     * @param removeTop number of tasks to pop from the last backup.
     * @param removeBottom number of tasks to pop from the BOTTOM of the last
     *        backup.
     * @param backup the tasks, as well as the partial result to add to the
     *        old backup.
     * @param log the restore-log.
     * @param thieves a list of thieves
     */
    public def this(val tan:Long
                    , val sourceId:Long
                    , val kind:Long
                    , val cause:Long
                    , val removeTop:Long
                    , val removeBottom:Long
                    , val log:Long
                    , val backup:Queue
                    , val thieves:List[Long]) : BackupData[Queue, R]
    {
        this.tan = tan;
        this.sourceId = sourceId;
        this.kind = kind;
        this.cause = cause;
        this.removeTop = removeTop;
        this.removeBottom = removeBottom;
        this.log = log;
        this.backup = backup;

        if (null == thieves)
        {
            this.thieves = new ArrayList[Long]();
        }
        else
        {
            this.thieves = thieves;
        }
    }
}