package de.uks.plm.ftglb.incremental.managers;

import x10.compiler.Inline;
import x10.compiler.Uncounted;

import x10.util.ArrayList;
import x10.util.List;
import x10.util.Pair;

import de.uks.plm.ftglb.incremental.TaskBag;
import de.uks.plm.ftglb.incremental.TaskQueue;
import de.uks.plm.ftglb.incremental.Worker;

import de.uks.plm.ftglb.incremental.dataStructures.BackupData;

import de.uks.plm.utils.AssertPrinter;

/**
 * This class manages the sending- and protocolling-process of backup-data.
 * For sending, a reference to the corresponding
 * @link{de.uts.plm.ftglb.incremental.Worker worker} is needed.
 */
public class BackupSenderManager[Queue, R] {Queue <: TaskQueue[Queue, R]}
{
    /**
     * Reference to the corresponding
     * @link{de.uts.plm.ftglb.incremental.Worker worker}.
     */
    public val worker:Worker[Queue, R];

    /**
     * The uppermost task of the min-state. Before sending, all lower tasks
     * are appended to this queue. Since this is a queue, the result at this
     * point in time is stored as well.
     */
    private var minState:Queue;

    /**
     * The size of the stored min-state. Note that this is the size without
     * the uppermost task.
     * @link{#minState minState}
     * @link{de.uts.plm.ftglb.incremental.TaskQueue#size() .size()}.
     */
    private var minSize:Long; 

    /**
     * This attribute hold the size of the last written backup.
     */
    private var backupSize:Long;

    public var log:Long;

    public def this(val worker:Worker[Queue, R]) : BackupSenderManager[Queue, R]
    {
        this.worker = worker;
        this.minSize = 0;
        this.backupSize = 0;
        this.minState = (null as Queue);
        this.log = this.worker.me;
    }

    /**
     * Protocol the current size of the queue as newS as well as the uppermost
     * task as theATask.
     * caller: local only.
     * 
     * active while in emergency: no.
     */
    // @Inline public def prot() : void
    // {
    //     if (0 == this.worker.queue.size())
    //     {
    //         this.minState = this.worker.queue.clone();
    //         this.minSize = -1;
    //     }
    //     else
    //     {
    //         this.minState = this.worker.queue.getTopElement();
    //         this.minSize = this.worker.queue.size() - 1;
    //     }
    //     // AssertPrinter.println(  "[Sender " + here + "]: new min-state of size "
    //     //                       + this.minSize + ".");
    // }

    // /**
    //  * Construct the complete queue to send as backup based on the current min-
    //  * state.
    //  * Sink: yes
    //  */
    // @Inline public def constructQueue() : Queue
    // {
    //     return (this.constructQueue(0));
    // }

    /**
     * Construct the complete queue to send as backup based on the current min-
     * state, as well as some tasks to remove from bottom (which got stolen 
     * away)
     * Sink: yes
     */
//     @Inline public def constructQueue () : Queue
//     {
//         assert (this.minState.size() <= 1) : here + ": minState has size > 1!";
//         val queueToSend:Queue = this.minState.clone();
//         if (0 != this.worker.queue.size())
//         {
//             val rest:Queue = this.worker.queue.getQueueFromBottom(this.minSize);
//             queueToSend.mergeAtBottom(rest);
//         }
// 
//         assert (queueToSend.size() == this.minSize + 1)
//                : here + ": constructed queue has wrong size!";
// 
//         return (queueToSend);
//     }

    /**
     * This method constructs a
     * @link{de.uts.plm.ftglb.incremental.dataStructures.BackupData
     * BackupData[Queue, R]}-object.
     *
     * @return the @link{de.uts.plm.ftglb.incremental.dataStructures.BackupData
     * BackupData[Queue, R]}-object corresponding to @link{#minState minState}
     * and @link{#minSize minSize}.
     */
//     @Inline public def getIncrementalData(val backupId:Long)
//         : BackupData[Queue, R]
//     {
//         val me:Long = this.worker.me;
//         var popTop:Long = this.worker.queue.size() - minSize - 1;
//         val rest:Queue = this.worker.queue.getQueueFromTop(this.minSize);
//         val queueToSend = this.minState.clone();
//         queueToSend.mergeAtBottom(rest);
// 
//         val data:BackupData[Queue, R]
//             = new BackupData[Queue, R](me, popTop, queueToSend);
// 
//         return (data);
//     }

    /**
     * Get the size of the minimal protocolled state
     */
    @Inline public def minSize() : Long
    {
        return (this.minSize);
    }

    /**
     * Send the backup-data to back.
     * Sink: no, we send the data to back.
     */
    @Inline private def send(val backupId:Long, val data:BackupData[Queue, R])
        : void
    {
        /*send
         * Encapsulate the st in a local val, so not the whole worker-object
         * gets send over the network.
         */
        val st:PlaceLocalHandle[Worker[Queue, R]] = this.worker.st;

        AssertPrinter.println(  "[Sender " + here + "]: sending backup to " 
                              + backupId + ", MinSize = " + (this.minSize + 1)
                              + ", RemoveTop = " + data.removeTop
                              + ", removeBottom = " + data.removeBottom
                              + ", backupSize = " + data.backup.size());

        val backup:Place = Place(backupId);
        at (backup) async
        {
            st().receiver.update(data);
        }
        this.backupSize = this.minSize + 1;

        AssertPrinter.println(  "[Sender " + here + "]: backupSize is now "
                              + this.backupSize);
    }
    /**
     * Create minimal recorded
     * @link{de.uks.plm.ftglb.incremental.TaskQueue queue}, pack it into an
     * @link{de.uks.plm.ftglb.incremental.dataStructures.BackupData 
     * BackupData[Queue, R]}-object and send it to the place, which backs
     * this place up.
     */
//     @Inline public def sendIncrementalBackupRegular(val backupId:Long) : void
//     {
//         if (this.worker.me == backupId)
//         {
//             this.worker.regAck(backupId);
//             return;
//         }
//         val removeTop:Long = Math.max(this.backupSize - Math.max(this.minSize
//                                                         , 0)
//                                       , 0);
// 
//         var queueToSend:Queue = this.minState.clone();
//         if (this.minSize > this.backupSize)
//         {
//             queueToSend = this.constructQueue();
// 
//             assert (queueToSend.size() == this.minSize + 1)
//                    : here + ": backup has wrong size of " + queueToSend.size() + " (should have " + (this.minSize + 1) + ").";
// 
//             val stableTasks:Long = Math.max(this.backupSize - 1, 0);
//             queueToSend.removeFromBottom(stableTasks);
// 
//             assert (queueToSend.size() + stableTasks == this.minSize + 1)
//                     : here + ": after removeBottom, the size of sending queue is wrong.";
//         }
// 
//         val me = this.worker.me;
// 
//         val data:BackupData[Queue, R]
//             = new BackupData[Queue, R](me, removeTop, queueToSend);
// 
//         AssertPrinter.println(  here + ": SendingIncrementalBackup, size "
//                               + queueToSend.size());
// 
//         this.send(backupId, data);
//         this.worker.resetK();
//     }

    /**
     * This method constructs and sends a full backup. Note, that the full
     * @link{de.uts.ftglb.incremental.Worker#queue worker.queue} at this point
     * in time is recorded in @link{#minState} and @link{minsize} and sent to
     * the backup-place.
     */
    @Inline public def sendFullBackupRegular(val backupId:Long) : void
    {
        if (this.worker.me == backupId)
        {
            this.worker.regAck(backupId);
            this.worker.state.afterMerge = false;
            this.worker.resetK();
            return;
        }
        // val queueToSend:Queue = this.constructQueue();
        val data:BackupData[Queue, R]
            = new BackupData[Queue, R](this.worker.me, this.worker.queue);

        AssertPrinter.println("[Sender " + here + "]: sending full, regular backup.");
        this.send(backupId, data);
        this.worker.state.afterMerge = false;
        this.worker.resetK();
    }

    /**
     * Send a full steal-backup
     * sink: no, indirectly switching places via send.
     */
    @Inline private def sendFullBackupSteal(val tan:Long
                                           , val backupId:Long
                                           , val removeBottom:Long
                                           , val thieves:List[Long]) : void
    {
        if (this.worker.me == backupId)
        {
            this.worker.stlAck(tan, backupId);
            return;
        }
        // val queueToSend:Queue = this.worker.queue;

        AssertPrinter.println(  here + ": sending full-steal backup of size "
                              + this.worker.queue.size() + ".");

        val data:BackupData[Queue, R]
            = new BackupData[Queue, R](tan
                                       , this.worker.me
                                       , this.worker.queue
                                       , thieves);

        this.send(backupId, data);
    }

    /**
     * This method acts for STLNreq and STLTreq. Depending on stolen, the
     * Backup-manager deciedes, which kind of backup to send and constructs
     * the backup-data accordingly.
     *
     * Important: this method DOES NOT CHECK, whether we can send. This must be
     * done before calling this method.
     *
     * TODO: maybe we can simplify this.
     */
    @Inline public def sendStealBackup(val tan:Long
                                       , val backupId:Long
                                       , val stolen:Long
                                       , val thieves:List[Long]) : void
    {
        /**
         * 1st case: we have not merged anything since the last backup and the
         * amount of stolen tasks is smaller than the protocolled min-state and
         * smaller than the last backup -> Simply send the number of stolen
         * tasks to the backup place. Do not reset the cycle.
         */
        if (this.worker.me == backupId)
        {
            this.worker.stlAck(tan, backupId);
            return;
        }
        // if (   (stolen <= backupSize)
        //     && (stolen <= this.minSize)
        //     && !this.worker.state.afterMerge)
        // {
        //     AssertPrinter.println("[Sender " + here + "]: case 1");
        //     val backup:BackupData[Queue,R]
        //         = new BackupData[Queue, R](tan
        //                                    , this.worker.me
        //                                    , BackupData.KIND_INCR
        //                                    , BackupData.CAUSE_STEAL
        //                                    , 0
        //                                    , stolen
        //                                    , -1
        //                                    , (null as Queue)
        //                                    , thieves);
        //     this.backupSize -= stolen;
        //     this.minSize -= stolen;
        //     this.send(backupId, backup);
        // }
        /**
         * 2nd cast: no merge since the last backup and the number of stolen 
         * tasks is bigger than the last backup, but smaller than the current 
         * min-state -> send the current min-state without the stolen tasks, but
         * do NOT reset the cycle. Basically a full backup.
         * 5-a)th case: We had a merge after the last backup, but the amount of
         * stolen tasks is smaller than the protocolled mind-state -> same
         * reaction.
         */
        // else if (   (stolen > backupSize && stolen <= this.minSize)
        //          || (this.worker.state.afterMerge && (stolen <= this.minSize)))
        // {
        //     AssertPrinter.println("[Sender " + here + "]: case 2");
        //     this.minSize -= stolen;
        //     this.sendFullBackupSteal(tan, backupId, stolen, thieves);
        // }
        /**
         * 3rd and 4th case: no merge since the last backup, but the number
         * of stolen tasks is bigger than the current min-state -> write 
         * full backup, reset cycle.
         * 5-bth case: merge since last backup and and the number of stolen
         * tasks is bigger than current min-state -> write full backup, reset
         * cycle
         */
        // else // if (stolen > this.minSize)
        // {
           AssertPrinter.println("[Sender " + here + "]: case 3");
           // this.prot();
            this.sendFullBackupSteal(tan, backupId, 0, thieves);
            this.worker.resetK();
        // }
    }

    /**
     * Send a full backup
     * sink: no, indirectly via send
     */
    @Inline private def sendFullBackup(val backupId:Long, val cause:Long) : void
    {
        // val queueToSend = this.constructQueue();

        var tan:Long = Long.MAX_VALUE;
        var thieves:List[Long] = null;
        if (this.worker.state.repeatStlReq)
        {
            tan = this.worker.currentTan;
            thieves = new ArrayList[Long]();
            for (entry:Pair[Long, TaskBag] in this.worker.tanDescriptor)
            {
                thieves.add(entry.first);
            }
        }
        val backup:BackupData[Queue, R]
            = new BackupData[Queue, R] (tan
                                        , this.worker.me
                                        , BackupData.KIND_COMP
                                        , cause
                                        , 0
                                        , 0
                                        , this.log
                                        , this.worker.queue
                                        , thieves);
        this.send(backupId, backup);
    }

    /**
     * This method functions as IA-Backup.
     * sink: no, indirectly via send
     */
    @Inline public def sendFullBackupIA(val backupId:Long) : void
    {
        if (this.worker.me == backupId)
        {
            this.worker.toAck(backupId);
            return;
        }
        this.sendFullBackup(backupId, BackupData.CAUSE_IA);
    }

    /**
     * This method functions as TO-Backup
     * sink: no, indirectly via send
     */
    @Inline public def sendFullBackupTO(val backupId:Long)
    {
        if (this.worker.me == backupId)
        {
            this.worker.toAck(backupId);
            return;
        }
        val tan:Long = Math.min(this.worker.currentTan, Long.MAX_VALUE);
        var thieves:List[Long] = new ArrayList[Long]();
        val tanDescriptor:List[Pair[Long, TaskBag]] = this.worker.tanDescriptor;
        for (entry:Pair[Long, TaskBag] in tanDescriptor)
        {
            thieves.add(entry.first);
        }
        if (thieves.isEmpty())
        {
            thieves = null;
        }

        val backup:BackupData[Queue, R]
            = new BackupData[Queue, R](tan
                                       , this.worker.me
                                       , BackupData.KIND_COMP
                                       , BackupData.CAUSE_TAKEOVER
                                       , 0
                                       , 0
                                       , -1
                                       , this.worker.queue
                                       , thieves);

        AssertPrinter.println(  "[Sender " + here +  "]: sending TO-Backup to "
                              + Place(backupId) + ". Thieves = " + thieves);

        this.send(backupId, backup);
    }
}