package de.uks.plm.ftglb.incremental.managers;

import x10.compiler.Inline;
import x10.compiler.Volatile;

import x10.util.ArrayList;
import x10.util.List;
import x10.util.RailUtils;

/**
 * This class manages the states of the local worker. All state-relevant 
 * variable are stored here.
 */
public class StateManager
{
    public static val REDUCER:(Boolean, Boolean) => Boolean
                      = (val b1:Boolean, val b2:Boolean) => (b1 || b2);

    @Volatile transient public var active:Boolean;
    @Volatile transient public var afterMerge:Boolean;
    @Volatile transient public var critical:Boolean;
    @Volatile transient public var emergency:Boolean;
    @Volatile transient public var empty:Boolean;
    @Volatile transient public var ghost:Boolean;
    @Volatile transient public var nextBackupFull:Boolean;
    @Volatile transient public var sendingBackup:Boolean;
    @Volatile transient public var stealFailed:Boolean;
    @Volatile transient public var valid:Boolean;
    @Volatile transient public var waiting:Boolean;
    @Volatile transient public var monitoring:Boolean;
    @Volatile transient public var sendBackupTO:Boolean;
    @Volatile transient public var restoreInProcess:Long;

    @Volatile transient public var answerRestore:Long;
    @Volatile transient public var answerRstAckAfterIAAck:Boolean;
    @Volatile transient public var blockTOBackup:Boolean;
    @Volatile transient public var repeatTOBackup:Boolean;
    @Volatile transient public var repeatStlReq:Boolean;

    @Inline public def this() : StateManager
    {
        this.active = false;
        this.afterMerge = false;
        this.critical = false;
        this.emergency = false;
        this.empty = true;
        this.ghost = false;
        this.nextBackupFull = false;
        this.sendingBackup = false;
        this.stealFailed = false;
        this.valid = true;
        this.waiting = false;
        this.monitoring = false;
        this.sendBackupTO = false;
        this.restoreInProcess = -1;
        this.answerRestore = -1;
        this.answerRstAckAfterIAAck = false;
        this.blockTOBackup = false;
        this.repeatTOBackup = false;
        this.repeatStlReq = false;
    }

    @Inline public def canSendBackup() : Boolean
    {
        return (!(this.sendingBackup || this.emergency));
    }
}